#include <linux/module.h>
#include <linux/init.h>
#include <linux/kernel.h>
#include <linux/fs.h>
#include <linux/cdev.h>
#include <asm/io.h>
#include <asm/uaccess.h>
#include <linux/ioport.h>
#include <linux/interrupt.h>
#include <linux/spinlock.h>
#include <linux/slab.h>

MODULE_DESCRIPTION("Keyboard hook");
MODULE_AUTHOR("SO2");
MODULE_LICENSE("GPL");

#define LOG_LEVEL		KERN_ALERT
#define MODULE_NAME		"keyboard_hook"

#define I8042_KBD_IRQ		1
#define I8042_STATUS_REG	0x64
#define I8042_DATA_REG		0x60

#define BUFFER_SIZE		1024
#define SCANCODE_RELEASED_MASK	0x80

static const char *placeholder_string = "keyboard_hook";

/**
 * is_key_press() -- checks if a scancode corresponds to a key press or releaase
 *
 * scancode - value read from the DATA register
 * returns non-zero for a key press and 0 otherwise
 */
static int is_key_press(int scancode)
{
	return !(scancode & SCANCODE_RELEASED_MASK);
}

/**
 * get_ascii() -- returns the character of the given scancode
 * Only works for alfanumeric/space/enter; returns '?' for other chars.
 *
 * scancode - value read from the DATA register
 */
static char get_ascii(int scancode)
{
	static char *row1 = "1234567890";
	static char *row2 = "qwertyuiop";
	static char *row3 = "asdfghjkl";
	static char *row4 = "zxcvbnm";

	scancode &= ~SCANCODE_RELEASED_MASK;
	if (scancode >= 0x02 && scancode <= 0x0b)
		return *(row1 + scancode - 0x02);
	if (scancode >= 0x10 && scancode <= 0x19)
		return *(row2 + scancode - 0x10);
	if (scancode >= 0x1e && scancode <= 0x26)
		return *(row3 + scancode - 0x1e);
	if (scancode >= 0x2c && scancode <= 0x32)
		return *(row4 + scancode - 0x2c);
	if (scancode == 0x39)
		return ' ';
	if (scancode == 0x1c)
		return '\n';
	return '?';
}

/**
 * TODO: Implement function to read byte from DATA register.
 * i8042_read_data() -- returns the value of the DATA register
 */
static inline u8 i8042_read_data(void)
{
	/* TODO: Implement function. */
}

/**
 * kbd_interrupt_handle() -- interrupt handler for IRQ 1 (keyboard IRQ)
 */
irqreturn_t kbd_interrupt_handle(int irq_no, void *dev_id)
{
	printk(LOG_LEVEL "IRQ %d\n", irq_no);

	return IRQ_NONE;
}

static int kbd_hook_init(void)
{
	int err;

	/* request IRQ */
	err = request_irq(I8042_KBD_IRQ,
			kbd_interrupt_handle,
			IRQF_SHARED, MODULE_NAME, (void *) placeholder_string);
	if (err != 0) {
		printk(LOG_LEVEL "ERROR: %s: error %d\n", "request_irq", err);
		return err;
	}

	return 0;
}

static void kbd_hook_exit(void)
{
	/* free IRQ */
	free_irq(I8042_KBD_IRQ, (void *) placeholder_string);
}

module_init(kbd_hook_init);
module_exit(kbd_hook_exit);

